/**
 * Class
 * @class EventBus
 */
export default class EventBus {
  constructor() {
    this.listeners = {};
    this.addEventListener = this.addEventListener.bind(this);
    this.removeEventListener = this.removeEventListener.bind(this);
    this.emit = this.emit.bind(this);
  }

  /**
   * make sure event name exists as a key in the listeners object
   * @param eventName name of the event
   * @private
   */
  _check(eventName) {
    if (!this.listeners[eventName]) {
      this.listeners[eventName] = [];
    }
  }

  /**
   * Setup an event listener
   * @param eventName Name of the event to listen to
   * @param listener Listener function
   * @param blocking Should the listener block the thread during its execution
   */
  addEventListener(eventName, listener, blocking = false) {
    this._check(eventName);
    this.listeners[eventName].push([listener, blocking]);
  }

  /**
   * Remove an event listener
   * @param eventName Name of the event listener was subscribed to
   * @param callback reference of the listener function
   */
  removeEventListener(eventName, callback) {
    this._check(eventName);
    this.listeners[eventName] = this.listeners[eventName].filter(
      ([f]) => f !== callback,
    );
  }

  /**
   * Event an event
   * @param eventName event to emit
   * @param args any kind of payload
   */
  emit(eventName, ...args) {
    this._check(eventName);
    this.listeners[eventName].forEach(([fn, blocking]) => {
      if (blocking) {
        fn(...args);
      } else {
        setTimeout(fn, 0, ...args);
      }
    });
  }
}
